/*
Copyright 1990-2003 Sun Microsystems, Inc. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions: The above copyright notice and this
permission notice shall be included in all copies or substantial
portions of the Software.


THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE OPEN GROUP OR SUN MICROSYSTEMS, INC. BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE EVEN IF
ADVISED IN ADVANCE OF THE POSSIBILITY OF SUCH DAMAGES.


Except as contained in this notice, the names of The Open Group and/or
Sun Microsystems, Inc. shall not be used in advertising or otherwise to
promote the sale, use or other dealings in this Software without prior
written authorization from The Open Group and/or Sun Microsystems,
Inc., as applicable.


X Window System is a trademark of The Open Group

OSF/1, OSF/Motif and Motif are registered trademarks, and OSF, the OSF
logo, LBX, X Window System, and Xinerama are trademarks of the Open
Group. All other trademarks and registered trademarks mentioned herein
are the property of their respective owners. No right, title or
interest in or to any trademark, service mark, logo or trade name of
Sun Microsystems, Inc. or its licensors is granted.

*/

#include <stdio.h>
#include "SunIM.h"
#include "hzinput.h"
#include "codetable.h"

/*********************** Key Mapping ***********************/

/*  get the keys that IME will use */
int map_keyevent_to_imekey(CodeTableStruct *hztbl, IMEKey key_event)
{
	int keycode = key_event->keyCode;
	int keychar = key_event->keyChar;
	int keystatus = key_event->modifier;

	DEBUG_printf("keycode: 0x%x, keychar:0x%x, keystatus: 0x%x\n", keycode, keychar, keystatus);

	/* Normal Alphameric key */
	if ((keycode >= IM_VK_COMMA && keycode <= IM_VK_CLOSE_BRACKET) ||
	    (keycode == IM_VK_QUOTE) || (keycode == IM_VK_BACK_QUOTE)) {
		if (! (keystatus & IM_CTRL_MASK) &&
		    ! (keystatus & IM_META_MASK) &&
		    ! (keystatus & IM_ALT_MASK) &&
		    ! (keystatus & IM_ALT_GRAPH_MASK) )
			return(keychar);
	}

	/* Special key */
	if (keystatus  == 0) {
		switch (keycode) {
			case IM_VK_ESCAPE:
				return(ESC_KEY);

			case IM_VK_BACK_SPACE:
				return(BACKSPACE_KEY);

			case IM_VK_ENTER:
				return(RETURN_KEY);

			case IM_VK_SPACE:
				return(SPACE_KEY);

			case IM_VK_DELETE:
				return(DELETE_KEY);

			case IM_VK_PAGE_UP:
				return(PAGEUP_KEY);

			case IM_VK_PAGE_DOWN:
				return(PAGEDOWN_KEY);
		}
		return(IME_NOT_USED_KEY);
	}

	/*  Ctrl status */
	if (keystatus == IM_CTRL_MASK) {
		if (keychar != 0) {
			/*  normal key  */
			unsigned char *keylist;
			int i, key_id;

			keychar = keycode | 0x80;

			key_id = -1;
			for (i = 0; i<MAX_FUNCTIONKEY_NUM; i++) {
				keylist = hztbl->functionkey[i].keylist;
				if (*keylist) {
					int j;
					int bDefined = 0;
					for (j=0; j<MAX_FUNCTIONKEY_LEN; j++) {
						if (keylist[j] == keychar) {
							bDefined = 1;
							break;
						}
					}
					if (bDefined) {
						key_id = i;
						break;
					}
				}
			}
			
			if (key_id != -1) {
				switch (key_id) {
					case PAGEUP_KEY_ID:
						return(PAGEUP_KEY);
					case PAGEDOWN_KEY_ID:
						return(PAGEDOWN_KEY);
					case BACKSPACE_KEY_ID:
						return(BACKSPACE_KEY);
					case CLEARALL_KEY_ID:
						return(ESC_KEY);
				}
			}
		}
	}

	return (IME_NOT_USED_KEY);
}

